// StaticTerm.cpp : implementation file
//

#include "stdafx.h"
#include "ilvideo.h"
#include "StaticTerm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define	ROP_DSPDxax	0x00E20746

/////////////////////////////////////////////////////////////////////////////
// CStaticTerm

IMPLEMENT_DYNAMIC(CStaticTerm, CStatic)

CStaticTerm::CStaticTerm(INT nCols,INT nRows,DWORD dwRefresh)
	: m_nCols(nCols)						// no. of visible columns
	, m_nRows(nRows)						// no. of visible rows
	, m_dwRefresh(dwRefresh)				// output refresh time in ms
{
	InitializeCriticalSection(&m_csLock);

	m_rgbTextColor = RGB(0xFF,0xFF,0xFF);	// white text color
	m_rgbBkColor   = RGB(0x00,0x00,0x00);	// black background color

	m_nBlinkFreq = 257;						// US blink frequency in ms

	m_nWidth = 0;							// character width
	m_nHeight = 0;							// character height

	m_xCurPos = 0; 							// x cursor position
	m_yCurPos = 0; 							// y cursor position

	m_hEvent = NULL;
	m_hWorkerThread = NULL;
	m_bCursorAct = false;					// showing text
	m_eCurState = CUR_OFF;					// cursor off

	// init insert cursor
	VERIFY(m_BmpInsCursor.LoadBitmap(IDB_INSERT));

	BITMAP CurBitMap;
	m_BmpInsCursor.GetBitmap(&CurBitMap);
	m_nInsCursorWidth = CurBitMap.bmWidth;
	m_nInsCursorHeight = CurBitMap.bmHeight;

	m_strBuffer = _T("");					// intermediate buffer
	m_hDrawThread = NULL;
	m_bRunning = false;
	m_hDrawEvent = NULL;					// event to start drawing timer

	StartDrawWorker();						// start worker thread for buffered output
}

CStaticTerm::~CStaticTerm()
{
	m_BmpInsCursor.DeleteObject();
	m_Font.DeleteObject();

	DeleteCriticalSection(&m_csLock);
}

BEGIN_MESSAGE_MAP(CStaticTerm, CStatic)
	//{{AFX_MSG_MAP(CStaticTerm)
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CStaticTerm message handlers

#if _MSC_VER <= 1200 // MSVC6 and earlier
BOOL CStaticTerm::OnChildNotify(UINT message, WPARAM wParam, LPARAM lParam, LRESULT* pLResult)
{
	// TODO: Add your specialized code here and/or call the base class
	if (message == WM_DRAWITEM)
	{
		ASSERT(pLResult == NULL);			// no return value expected
		UNUSED(pLResult);					// unused in release builds
		DrawItem(reinterpret_cast<LPDRAWITEMSTRUCT>(lParam));
		return TRUE;
	}
	return CStatic::OnChildNotify(message, wParam, lParam, pLResult);
}
#endif

void CStaticTerm::PreSubclassWindow()
{
	// TODO: Add your specialized code here and/or call the base class
	ModifyStyle(0, SS_OWNERDRAW);

	LOGFONT lf;
	// create a non antialiased font to disable ClearType
	GetFont()->GetLogFont(&lf);
	lf.lfQuality = NONANTIALIASED_QUALITY;
	m_Font.CreateFontIndirect(&lf);
	SetFont(&m_Font,FALSE);

	CStatic::PreSubclassWindow();
}

void CStaticTerm::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	// TODO: Add your message handler code here and/or call default
	CBitmap bmpText,*pbmpText;
	CBitmap bmpMask,*pbmpMask;
	CDC dcMask,dcText;
	INT nLinCurPos,nLen;
	bool bCurInv;
	TEXTMETRIC tm;

	// draw bit not set
	if ((lpDrawItemStruct->itemAction & ODA_DRAWENTIRE) == 0) return;

	CRect rc = lpDrawItemStruct->rcItem;
	CDC *pDC = CDC::FromHandle(lpDrawItemStruct->hDC);

	pDC->GetTextMetrics(&tm);				// get character dimensions
	m_nWidth = tm.tmAveCharWidth;
	m_nHeight = tm.tmHeight;

	// create text DC
	dcText.CreateCompatibleDC(pDC);
	VERIFY(bmpText.CreateCompatibleBitmap(pDC,rc.Width(),rc.Height()));
	pbmpText = dcText.SelectObject(&bmpText);

	// create mask DC
	dcMask.CreateCompatibleDC(pDC);
	VERIFY(bmpMask.CreateCompatibleBitmap(pDC,rc.Width(),rc.Height()));
	pbmpMask = dcMask.SelectObject(&bmpMask);

	// fill mask DC
	VERIFY(dcMask.PatBlt(0,0,rc.Width(),rc.Height(),BLACKNESS));

	nLinCurPos = (m_yCurPos * (m_nCols + 2)) + m_xCurPos;
	bCurInv = false;						// draw cursor not inverse

	nLen = GetWindowTextLength();			// get length of window content

	// buffer for the text content
	LPTSTR pcBuffer = static_cast<LPTSTR>(new TCHAR[nLen+1]);
	if (pcBuffer != NULL)
	{
		GetWindowText(pcBuffer,nLen+1);		// get item text

		// modify pcBuffer -> if Bit 7 is set, clear it and set the invert mask at this character
		for (INT i = 0; i < nLen; ++i)
		{
			if ((pcBuffer[i] & 0x80) != 0)	// show inverse
			{
				if (i == nLinCurPos)		// linear address of cursor
				{
					bCurInv = true;			// draw cursor inverse
				}

				pcBuffer[i] &= ~0x80;		// clear bit

				// set inverse mask
				VERIFY(dcMask.PatBlt((i % (m_nCols + 2)) * m_nWidth,
									 (i / (m_nCols + 2)) * m_nHeight,
									 m_nWidth,
									 m_nHeight,
									 WHITENESS));
			}
		}

		// fill text DC with B&W text
		dcText.SetTextColor(RGB(255,255,255));
		dcText.SetBkColor(RGB(0,0,0));

		CFont *pFontPrev = dcText.SelectObject(GetFont());
		dcText.DrawText(pcBuffer,nLen,&rc,DT_LEFT | DT_NOPREFIX | DT_NOCLIP);
		dcText.SelectObject(pFontPrev);

		delete[] pcBuffer;
	}

	// add text to the background invert mask
	VERIFY(dcMask.BitBlt(rc.left,
						 rc.top,
						 rc.Width(),
						 rc.Height(),
						 &dcText,
						 rc.left,
						 rc.top,
						 SRCINVERT));

	if (m_bCursorAct)						// show cursor
	{
		CDC     dcCur;
		CBitmap *pbmpCur;

		// select the correct cursor
		switch (m_eCurState)
		{
		case CUR_OFF:
			break;
		case CUR_REPLACE:
			VERIFY(dcMask.PatBlt(m_xCurPos * m_nWidth,
								 m_yCurPos * m_nHeight,
								 m_nWidth,
								 m_nHeight,
								 DSTINVERT));
			break;
		case CUR_INSERT:
			// write insert cursor to mask
			dcCur.CreateCompatibleDC(pDC);
			pbmpCur = dcCur.SelectObject(&m_BmpInsCursor);
			VERIFY(dcMask.StretchBlt(m_xCurPos * m_nWidth,
									 m_yCurPos * m_nHeight,
									 m_nWidth,
									 m_nHeight,
									 &dcCur,
									 0,
									 0,
									 m_nInsCursorWidth,
									 m_nInsCursorHeight,
									 (bCurInv) ? NOTSRCCOPY : SRCCOPY));
			dcCur.SelectObject(pbmpCur);
			break;
		default: ASSERT(false);
		}
	}

	// convert the BW to a two color bitmap
	dcText.FillSolidRect(rc,m_rgbBkColor);	// fill with background color

	CBrush Brush,*pBrush;
	Brush.CreateSolidBrush(m_rgbTextColor);	// foreground color brush
	pBrush = dcText.SelectObject(&Brush);
	VERIFY(dcText.BitBlt(rc.left,
						 rc.top,
						 rc.Width(),
						 rc.Height(),
						 &dcMask,
						 rc.left,
						 rc.top,
						 ROP_DSPDxax));
	dcText.SelectObject(pBrush);
	VERIFY(Brush.DeleteObject());

	// update display
	VERIFY(pDC->BitBlt(rc.left,
					   rc.top,
					   rc.Width(),
					   rc.Height(),
					   &dcText,
					   rc.left,
					   rc.top,
					   SRCCOPY));

	dcMask.SelectObject(pbmpMask);			// delete mask
	VERIFY(bmpMask.DeleteObject());

	dcText.SelectObject(pbmpText);			// delete text
	VERIFY(bmpText.DeleteObject());
}

void CStaticTerm::OnDestroy()
{
	CStatic::OnDestroy();

	// TODO: Add your message handler code here
	StopBlinkingThread();					// destroy cursor thread
	StopDrawWorker();						// stop the worker thread
}

VOID CStaticTerm::InvalidateCursor()
{
	RECT rc;

	rc.left   = m_xCurPos * m_nWidth;
	rc.top    = m_yCurPos * m_nHeight;
	rc.right  = rc.left + m_nWidth;
	rc.bottom = rc.top  + m_nHeight;
	InvalidateRect(&rc,FALSE);
	return;
}

UINT CStaticTerm::ThreadCusorBlink(LPVOID pParam)
{
	CStaticTerm *p = reinterpret_cast<CStaticTerm *>(pParam);
	ASSERT(p != NULL && p->IsKindOf(RUNTIME_CLASS(CStaticTerm)));

	// show cursor
	p->m_bCursorAct = true;					// show cursor
	p->InvalidateCursor();

	while (WaitForSingleObject(p->m_hEvent,p->m_nBlinkFreq) == WAIT_TIMEOUT)
	{
		p->m_bCursorAct = !p->m_bCursorAct;	// toogle cursor / normal char
		p->InvalidateCursor();
	}

	// force to exit with normal char
	p->m_bCursorAct = false;				// show text
	p->InvalidateCursor();
	return 0;
}

//
// start cursor blinking thread
//
bool CStaticTerm::StartBlinkingThread()
{
	DWORD dwThreadID;

	bool bSucc = false;

	if (m_hEvent == NULL && m_hWorkerThread == NULL)
	{
		// create shutdown event
		VERIFY(m_hEvent = CreateEvent(NULL,FALSE,FALSE,NULL));

		// create blinking thread
		m_hWorkerThread = CreateThread(NULL,
									   0,
									   (LPTHREAD_START_ROUTINE) ThreadCusorBlink,
									   (LPVOID) this,
									   0,
									   &dwThreadID);
		ASSERT(m_hWorkerThread);
		bSucc = (m_hWorkerThread != NULL);
	}
	return bSucc;
}

//
// stop cursor blinking thread
//
VOID CStaticTerm::StopBlinkingThread()
{
	if (m_hWorkerThread)					// shut down thread
	{
		SetEvent(m_hEvent); 				// set finish event

		// wait for thread down
		WaitForSingleObject(m_hWorkerThread,INFINITE);
		CloseHandle(m_hWorkerThread);
		m_hWorkerThread = NULL;
	}

	if (m_hEvent)							// close event
	{
		CloseHandle(m_hEvent);
		m_hEvent = NULL;
	}
	return;
}

//
// set new text and background color
//
VOID CStaticTerm::SetColors(COLORREF rgbTextColor, COLORREF rgbBkColor)
{
	m_rgbTextColor = rgbTextColor;			// text color
	m_rgbBkColor = rgbBkColor;				// background color
	Invalidate(FALSE);						// redraw all
	return;
}

//
// set new cursor position and state
//
VOID CStaticTerm::SetCursor(INT x, INT y, enum CUR_STATE eState)
{
	ASSERT(m_eCurState >= CUR_OFF && m_eCurState <= CUR_INSERT);
	ASSERT(eState      >= CUR_OFF && eState      <= CUR_INSERT);

	x %= m_nCols;							// limit coordinates to range
	y %= m_nRows;

	// cursor state or position change
	if (eState != m_eCurState || x != m_xCurPos || y != m_yCurPos)
	{
		m_eCurState = CUR_OFF;				// switch cursor off
		StopBlinkingThread();
		InvalidateCursor();					// redraw character under cursor
	}

	m_xCurPos = x;							// save new cursor position
	m_yCurPos = y;

	if (eState != CUR_OFF)					// switch on cursor
	{
		StartBlinkingThread();
	}
	m_eCurState = eState;					// save new state
	return;
}

//
// worker thread for drawing
//
VOID CStaticTerm::SetBufferText(LPCTSTR lpszText)
{
	EnterCriticalSection(&m_csLock);
	{
		BOOL bTrigger = m_strBuffer.IsEmpty();

		m_strBuffer = (LPCTSTR) lpszText;
		if (bTrigger)
		{
			SetEvent(m_hDrawEvent);			// start update timer
		}
	}
	LeaveCriticalSection(&m_csLock);
	return;
}

//
// flush buffered text
//
VOID CStaticTerm::FlushText()
{
	EnterCriticalSection(&m_csLock);
	{
		if (!m_strBuffer.IsEmpty())
		{
			// write buffer
			SetWindowText(m_strBuffer);
			m_strBuffer.Empty();
		}
	}
	LeaveCriticalSection(&m_csLock);
	return;
}

//
// worker thread for drawing
//
UINT CStaticTerm::DrawThread(LPVOID pParam)
{
	CStaticTerm *p = reinterpret_cast<CStaticTerm *>(pParam);

	DWORD dwTimeout = INFINITE;

	ASSERT(p != NULL);

	while (p->m_bRunning)
	{
		switch (WaitForSingleObject(p->m_hDrawEvent,dwTimeout))
		{
		case WAIT_TIMEOUT: // update display output
			p->FlushText();
			dwTimeout = INFINITE;
			break;
		case WAIT_OBJECT_0: // trigger output
			dwTimeout = p->m_dwRefresh;
			break;
		}
	}
	return 0;
}

//
// start drawing worker thread
//
bool CStaticTerm::StartDrawWorker()
{
	DWORD dwThreadID;

	bool bSucc = false;

	m_bRunning = true;

	if (m_hDrawEvent == NULL && m_hDrawThread == NULL)
	{
		// create drawing timer event
		VERIFY(m_hDrawEvent = CreateEvent(NULL,FALSE,FALSE,NULL));

		// create drawing thread
		m_hDrawThread = CreateThread(NULL,
									 0,
									 (LPTHREAD_START_ROUTINE) DrawThread,
									 (LPVOID) this,
									 0,
									 &dwThreadID);
		ASSERT(m_hDrawThread);

		bSucc = (m_hDrawThread != NULL);
	}
	return bSucc;
}

//
// shut down drawing worker thread
//
VOID CStaticTerm::StopDrawWorker()
{
	if (m_hDrawThread)						// worker thread running
	{
		m_bRunning = false;					// exit worker thread
		SetEvent(m_hDrawEvent);				// kill waiting thread

		// wait for worker thread down
		WaitForSingleObject(m_hDrawThread,INFINITE);
		CloseHandle(m_hDrawThread);
		m_hDrawThread = NULL;

		FlushText();						// write rest of buffer
	}

	if (m_hDrawEvent)						// close event
	{
		CloseHandle(m_hDrawEvent);
		m_hDrawEvent = NULL;
	}
	return;
}
