// Vid32Dev.cpp: implementation of the CVid32Dev class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Vid32Dev.h"

// address in display memory
#define CHR_AT(x,y)		(m_pbyBuffer[((m_nBufLine + (y)) % m_nBufRows) * m_nVisCols + (x)])

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CVid32Dev::CVid32Dev(CStaticTerm& wndScreen,INT nCols,INT nRows)
	: m_wndScreen(wndScreen),m_nVisCols(nCols),m_nVisRows(nRows),CIldevBase(false)
{
	// HP-IL data and variables
	m_byAID = 48;							// Accessory ID = video
	m_byDEFADDR = 31;						// default address after AAU
	m_strDID = _T("");						// Device ID (no response)

	// video data and variables
	m_nBufRows = 31;						// no. of buffer rows
	m_nBufLine = 0;							// buffer line of screen view

	m_xCurPos = 0;							// x cursor position
	m_yCurPos = 0;							// y cursor position

	m_nEsc = 0; 							// no escape sequence
	m_bAutoCrLf = false;					// reset automatic CR LF flag

	// video buffer
	VERIFY(m_pbyBuffer = static_cast<BYTE *>(new BYTE[m_nVisCols * m_nBufRows]));
	FillMemory(m_pbyBuffer,m_nVisCols * m_nBufRows,' ');

	SyncScreen();							// synchronize screen buffer

	m_bCursorShow = true;					// show cursor
	m_eCurState = CUR_REPLACE;				// cursor is replace cursur
	SetCursorPos(0,0);						// set cursor
}

CVid32Dev::~CVid32Dev()
{
	delete[] m_pbyBuffer;					// video buffer
}

//
// synchronize screen with display memory buffer
//
VOID CVid32Dev::SyncScreen()
{
	LPTSTR lpszText;
	INT    nAddr,nBuflen;

	nAddr = m_nBufLine * m_nVisCols;		// begin inside display memory

	// buffer length
	nBuflen = (m_nVisCols  + 2) * m_nVisRows + 1;

	// allocate output string buffer
	VERIFY(lpszText = static_cast<LPTSTR>(new TCHAR[nBuflen]));
	if (lpszText != NULL)					// buffer allocated
	{
		INT nLinAddr = 0;
		for (INT y = 0; y < m_nVisRows; ++y)
		{
			for (INT x = 0; x < m_nVisCols; ++x)
			{
				ASSERT(nLinAddr < nBuflen);
				ASSERT(nAddr < (m_nVisCols * m_nBufRows));
				lpszText[nLinAddr++] = static_cast<TCHAR>(m_pbyBuffer[nAddr++]);
				nAddr %= m_nVisCols * m_nBufRows; // address wrap around
			}
			lpszText[nLinAddr++] = _T('\r');
			lpszText[nLinAddr++] = _T('\n');
		}
		ASSERT(nLinAddr < nBuflen);
		lpszText[nLinAddr] = 0;				// EOS

		ASSERT(nLinAddr + 1 == nBuflen);	// text buffer has the expected length
		m_wndScreen.SetBufferText(lpszText);
		delete[] lpszText;
	}
	return;
}

//
// set new cursor position
//
VOID CVid32Dev::SetCursorPos(INT x, INT y)
{
	ASSERT (x >= 0 && y >= 0);

	x %= m_nVisCols;
	y %= m_nVisRows;

	// set cursor state
	m_wndScreen.SetCursor(x,y,m_bCursorShow ? m_eCurState : CUR_OFF);

	m_xCurPos = x;
	m_yCurPos = y;
	return;
}

//
// move cursor left
//
VOID CVid32Dev::CursorLeft()
{
	// nove cursor left
	if (m_xCurPos-- == 0)					// already at first column
	{
		m_xCurPos = m_nVisCols - 1; 		// goto last column

		// move cursor up
		if (m_yCurPos-- == 0)				// already at first row
		{
			m_xCurPos = m_yCurPos = 0;		// set cursur to (0,0)
		}
	}
	return;
}

//
// move cursor right
//
VOID CVid32Dev::CursorRight()
{
	if (m_xCurPos++ == m_nVisCols - 1)
	{
		m_xCurPos = 0;
		if (m_yCurPos++ == m_nVisRows - 1)
			m_yCurPos = 0;
	}
	return;
}

//
// eval line feed
//
VOID CVid32Dev::EvalLF()
{
	if (++m_yCurPos == m_nVisRows)			// end of screen
	{
		m_yCurPos = m_nVisRows - 1;			// set cursor to last line
		m_nBufLine = (m_nBufLine + 1) % m_nBufRows; // set new line where screen begin
		for (INT i = 0; i < m_nVisCols; ++i) // clear next line
		{
			CHR_AT(i,m_yCurPos) = ' ';
		}
		SyncScreen();						// synchronize screen buffer
	}
	return;
}

//
// Device Clear
// empty display & screen buffer and show replace cursor at (0,0)
//
VOID CVid32Dev::ClearDevice()
{
	// empty display buffer
	FillMemory(m_pbyBuffer,m_nVisCols * m_nBufRows,' ');
	m_nBufLine = 0; 						// view display buffer from top
	SyncScreen();							// synchronize screen buffer
	m_bCursorShow = true;					// show
	m_eCurState = CUR_REPLACE;				// replace
	SetCursorPos(0,0);						// cursor at (0,0)
	return;
}

//
// received data byte in listener mode
//
VOID CVid32Dev::InData(WORD wFrame)
{
	INT i,j;

	BYTE byData = static_cast<BYTE>(wFrame);

	// this is the virtual loop
	TRACE(_T("ILVideo Data = %02X\n"),byData);

	if (m_nEsc > 0)							// escape sequence
	{
		--m_nEsc;							// handled ESC sequence

		switch (m_nEsc)
		{
		case 0: 							// single ESC
			switch (byData)
			{
			case 'A': // Cursor Up
				if (m_yCurPos-- == 0)		// move cursor up
					m_yCurPos = 0;
				break;
			case 'B': // Cursor Down
				if (m_yCurPos++ == m_nVisRows - 1)
					m_yCurPos = m_nVisRows - 1;
				break;
			case 'C': // Cursor Right
				CursorRight();
				break;
			case 'D': // Cursor Left
				CursorLeft();
				break;
			case 'E': // Device Clear
				ClearDevice();
				break;
			case 'H': // Cursor Home
				m_xCurPos = m_yCurPos = 0;
				break;
			case 'J': // Clear Display From Cursor
				i = m_xCurPos;				// first clear until end of current line

				// then clear the lines after current position
				for (j = m_yCurPos; j < m_nVisRows; ++j)
				{
					for (; i < m_nVisCols; ++i)
					{
						CHR_AT(i,j) = ' ';
					}
					i = 0;					// clear all other lines from beginning
				}
				SyncScreen();				// synchronize screen buffer
				break;
			case 'Q': // Insert Cursor
				m_eCurState = CUR_INSERT;
				break;
			case 'R': // Replace Cursor
				m_eCurState = CUR_REPLACE;
				break;
			case 'S': // Roll Up
				if (++m_nBufLine == m_nBufRows)
					m_nBufLine = 0;
				SyncScreen();				// synchronize screen buffer
				break;
			case 'T': // Roll Down
				if (m_nBufLine-- == 0)
					m_nBufLine = m_nBufRows - 1;
				SyncScreen();				// synchronize screen buffer
				break;
			case '<': // Cursor Off
				m_bCursorShow = false;
				break;
			case '>': // Cursor On
				m_bCursorShow = true;
				break;
			case '%': // Cursor To Address
				m_nEsc = 3;					// still need two arguments, fetch column next
				break;
			default: // illegal ESC sequence
				break;						// ignore sequence
			}
			break;

		case 2: // Cursor To Address, m column argument
			m_xCurPos = byData % m_nVisCols;
			m_nEsc = 2;						// still need one argument, fetch row next
			break;

		case 1:  // Cursor To Address, n row argument
			m_yCurPos = byData % m_nVisRows;
			m_nEsc = 0;						// handled ESC sequence
			break;

		default:
			ASSERT(false);
		}

		m_bAutoCrLf = false;				// reset automatic CR LF flag
	}
	else
	{
		switch (byData)
		{
		case 8:  // BS
			CursorLeft();					// move cursor left
			m_bAutoCrLf = false;			// reset automatic CR LF flag
			break;
		case 10: // LF
			if (!m_bAutoCrLf)				// not evaluated automatic CR LF
			{
				EvalLF();					// eval LF
			}
			m_bAutoCrLf = false;			// reset automatic CR LF flag
			break;
		case 13: // CR
			m_xCurPos = 0;					// cursor to column 0
			break;
		case 27: // ESC
			m_nEsc = 1; 					// escape sequence
			break;
		default:
			// displayable character?
			if ((byData & 0x7F) >= 32 && (byData & 0x7F) <= 126)
			{
				m_bAutoCrLf = false;		// reset automatic CR LF flag

				CHR_AT(m_xCurPos,m_yCurPos) = byData;
				SyncScreen();				// synchronize screen buffer

				// end of column?
				if (++m_xCurPos == m_nVisCols)
				{
					m_xCurPos = 0;			// eval CR
					EvalLF();				// eval LF
					m_bAutoCrLf = true;		// evaluated automatic CR LF
				}
			}
			break;
		}
	}

	SetCursorPos(m_xCurPos,m_yCurPos);		// update cursor
}
